@@ -5,7 +5,7 @@
 //
 // Author: Mike McCauley
 // Copyright (C) 2011-2013 Mike McCauley
-// $Id: bcm2835.c,v 1.23 2015/03/31 04:55:41 mikem Exp mikem $
+// $Id: bcm2835.c,v 1.26 2018/08/27 20:45:57 mikem Exp mikem $
 */
 
 
@@ -18,7 +18,6 @@
 #include <time.h>
 #include <unistd.h>
 #include <sys/types.h>
-#include <sys/utsname.h>
 
 #define BCK2835_LIBRARY_BUILD
 #include "bcm2835.h"
@@ -56,18 +55,77 @@ volatile uint32_t *bcm2835_spi0        = (uint32_t *)MAP_FAILED;
 volatile uint32_t *bcm2835_bsc0        = (uint32_t *)MAP_FAILED;
 volatile uint32_t *bcm2835_bsc1        = (uint32_t *)MAP_FAILED;
 volatile uint32_t *bcm2835_st	       = (uint32_t *)MAP_FAILED;
+volatile uint32_t *bcm2835_aux	       = (uint32_t *)MAP_FAILED;
+volatile uint32_t *bcm2835_spi1        = (uint32_t *)MAP_FAILED;
+
 
 
 /* This variable allows us to test on hardware other than RPi.
 // It prevents access to the kernel memory, and does not do any peripheral access
 // Instead it prints out what it _would_ do if debug were 0
-*/
+ */
 static uint8_t debug = 0;
 
+/* RPI 4 has different pullup registers - we need to know if we have that type */
+
+static uint8_t pud_type_rpi4 = 0;
+
+/* RPI 4 has different pullup operation - make backwards compat */
+
+static uint8_t pud_compat_setting = BCM2835_GPIO_PUD_OFF;
+
 /* I2C The time needed to transmit one byte. In microseconds.
  */
 static int i2c_byte_wait_us = 0;
 
+/* SPI bit order. BCM2835 SPI0 only supports MSBFIRST, so we instead 
+ * have a software based bit reversal, based on a contribution by Damiano Benedetti
+ */
+static uint8_t bcm2835_spi_bit_order = BCM2835_SPI_BIT_ORDER_MSBFIRST;
+static uint8_t bcm2835_byte_reverse_table[] = 
+{
+    0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
+    0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
+    0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
+    0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
+    0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
+    0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
+    0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
+    0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
+    0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
+    0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
+    0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
+    0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
+    0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
+    0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
+    0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
+    0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
+    0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
+    0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
+    0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
+    0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
+    0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
+    0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
+    0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
+    0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
+    0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
+    0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
+    0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
+    0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
+    0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
+    0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
+    0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
+    0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
+};
+
+static uint8_t bcm2835_correct_order(uint8_t b)
+{
+    if (bcm2835_spi_bit_order == BCM2835_SPI_BIT_ORDER_LSBFIRST)
+	return bcm2835_byte_reverse_table[b];
+    else
+	return b;
+}
+
 /*
 // Low level register access functions
 */
@@ -93,6 +151,11 @@ uint32_t* bcm2835_regbase(uint8_t regbase)
 	    return (uint32_t *)bcm2835_bsc0;
 	case BCM2835_REGBASE_BSC1:
 	    return (uint32_t *)bcm2835_st;
+	case BCM2835_REGBASE_AUX:
+	    return (uint32_t *)bcm2835_aux;
+	case BCM2835_REGBASE_SPI1:
+	    return (uint32_t *)bcm2835_spi1;
+
     }
     return (uint32_t *)MAP_FAILED;
 }
@@ -115,8 +178,8 @@ uint32_t bcm2835_peri_read(volatile uint32_t* paddr)
     uint32_t ret;
     if (debug)
     {
-        printf("bcm2835_peri_read  paddr %08X\n", (unsigned) paddr);
-	return 0;
+		printf("bcm2835_peri_read  paddr %p\n", (void *) paddr);
+		return 0;
     }
     else
     {
@@ -137,7 +200,7 @@ uint32_t bcm2835_peri_read_nb(volatile uint32_t* paddr)
 {
     if (debug)
     {
-	printf("bcm2835_peri_read_nb  paddr %08X\n", (unsigned) paddr);
+	printf("bcm2835_peri_read_nb  paddr %p\n", paddr);
 	return 0;
     }
     else
@@ -153,7 +216,7 @@ void bcm2835_peri_write(volatile uint32_t* paddr, uint32_t value)
 {
     if (debug)
     {
-	printf("bcm2835_peri_write paddr %08X, value %08X\n", (unsigned) paddr, value);
+	printf("bcm2835_peri_write paddr %p, value %08X\n", paddr, value);
     }
     else
     {
@@ -168,8 +231,8 @@ void bcm2835_peri_write_nb(volatile uint32_t* paddr, uint32_t value)
 {
     if (debug)
     {
-	printf("bcm2835_peri_write_nb paddr %08X, value %08X\n",
-               (unsigned) paddr, value);
+	printf("bcm2835_peri_write_nb paddr %p, value %08X\n",
+                paddr, value);
     }
     else
     {
@@ -390,8 +453,14 @@ void bcm2835_gpio_clr_afen(uint8_t pin)
 /* Set pullup/down */
 void bcm2835_gpio_pud(uint8_t pud)
 {
-    volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUD/4;
-    bcm2835_peri_write(paddr, pud);
+    if( pud_type_rpi4 )
+    {
+        pud_compat_setting = pud;
+    }
+    else {
+        volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUD/4;
+        bcm2835_peri_write(paddr, pud);
+    }
 }
 
 /* Pullup/down clock
@@ -399,9 +468,17 @@ void bcm2835_gpio_pud(uint8_t pud)
 */
 void bcm2835_gpio_pudclk(uint8_t pin, uint8_t on)
 {
-    volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUDCLK0/4 + pin/32;
-    uint8_t shift = pin % 32;
-    bcm2835_peri_write(paddr, (on ? 1 : 0) << shift);
+    if( pud_type_rpi4 )
+    {
+        if( on )
+            bcm2835_gpio_set_pud( pin, pud_compat_setting);
+    }
+    else
+    {
+        volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUDCLK0/4 + pin/32;
+        uint8_t shift = pin % 32;
+        bcm2835_peri_write(paddr, (on ? 1 : 0) << shift);
+    }
 }
 
 /* Read GPIO pad behaviour for groups of GPIOs */
@@ -448,7 +525,7 @@ void bcm2835_delayMicroseconds(uint64_t micros)
     if (debug)
     {
 	/* Cant access sytem timers in debug mode */
-	printf("bcm2835_delayMicroseconds %lld\n", micros);
+	printf("bcm2835_delayMicroseconds %lld\n", (long long int) micros);
 	return;
     }
 
@@ -522,17 +599,70 @@ void bcm2835_gpio_write_mask(uint32_t value, uint32_t mask)
 // 6. Write to GPPUDCLK0/1 to remove the clock
 //
 // RPi has P1-03 and P1-05 with 1k8 pullup resistor
+//
+// RPI 4 uses a different PUD method - no clock
+
 */
 void bcm2835_gpio_set_pud(uint8_t pin, uint8_t pud)
 {
-    bcm2835_gpio_pud(pud);
-    delayMicroseconds(10);
-    bcm2835_gpio_pudclk(pin, 1);
-    delayMicroseconds(10);
-    bcm2835_gpio_pud(BCM2835_GPIO_PUD_OFF);
-    bcm2835_gpio_pudclk(pin, 0);
+    if( pud_type_rpi4 )
+    {
+        int shiftbits = (pin & 0xf) << 1;
+        uint32_t bits;
+        uint32_t pull;
+        
+        switch (pud)
+        {
+           case BCM2835_GPIO_PUD_OFF:  pull = 0; break;
+           case BCM2835_GPIO_PUD_UP:   pull = 1; break;
+           case BCM2835_GPIO_PUD_DOWN: pull = 2; break;
+           default: return;
+        }
+                
+        volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUPPDN0/4 + (pin >> 4);
+        
+        bits = bcm2835_peri_read_nb( paddr );
+        bits &= ~(3 << shiftbits);
+        bits |= (pull << shiftbits);
+        
+        bcm2835_peri_write_nb( paddr, bits );
+        
+    } else
+    {
+        bcm2835_gpio_pud(pud);
+        delayMicroseconds(10);
+        bcm2835_gpio_pudclk(pin, 1);
+        delayMicroseconds(10);
+        bcm2835_gpio_pud(BCM2835_GPIO_PUD_OFF);
+        bcm2835_gpio_pudclk(pin, 0);
+    }
+    
 }
 
+
+uint8_t bcm2835_gpio_get_pud(uint8_t pin)
+{
+    uint8_t ret = BCM2835_GPIO_PUD_ERROR;
+    
+    if( pud_type_rpi4 )
+    {
+        uint32_t bits;
+        volatile uint32_t* paddr = bcm2835_gpio + BCM2835_GPPUPPDN0/4 + (pin >> 4);
+        bits = (bcm2835_peri_read_nb( paddr ) >> ((pin & 0xf)<<1)) & 0x3;
+        
+        switch (bits)
+        {
+            case 0: ret = BCM2835_GPIO_PUD_OFF; break;
+            case 1: ret = BCM2835_GPIO_PUD_UP; break;
+            case 2: ret = BCM2835_GPIO_PUD_DOWN; break;
+            default: ret = BCM2835_GPIO_PUD_ERROR;
+        }   
+    }
+    
+    return ret;
+}
+
+
 int bcm2835_spi_begin(void)
 {
     volatile uint32_t* paddr;
@@ -567,9 +697,9 @@ void bcm2835_spi_end(void)
     bcm2835_gpio_fsel(RPI_GPIO_P1_23, BCM2835_GPIO_FSEL_INPT); /* CLK */
 }
 
-void bcm2835_spi_setBitOrder(uint8_t __attribute__((unused)) order)
+void bcm2835_spi_setBitOrder(uint8_t order)
 {
-    /* BCM2835_SPI_BIT_ORDER_MSBFIRST is the only one supported by SPI0 */
+    bcm2835_spi_bit_order = order;
 }
 
 /* defaults to 0, which means a divider of 65536.
@@ -583,6 +713,13 @@ void bcm2835_spi_setClockDivider(uint16_t divider)
     bcm2835_peri_write(paddr, divider);
 }
 
+void bcm2835_spi_set_speed_hz(uint32_t speed_hz)
+{
+	uint16_t divider = (uint16_t) ((uint32_t) BCM2835_CORE_CLK_HZ / speed_hz);
+	divider &= 0xFFFE;
+	bcm2835_spi_setClockDivider(divider);
+}
+
 void bcm2835_spi_setDataMode(uint8_t mode)
 {
     volatile uint32_t* paddr = bcm2835_spi0 + BCM2835_SPI0_CS/4;
@@ -612,14 +749,14 @@ uint8_t bcm2835_spi_transfer(uint8_t value)
 	;
 
     /* Write to FIFO, no barrier */
-    bcm2835_peri_write_nb(fifo, value);
+    bcm2835_peri_write_nb(fifo, bcm2835_correct_order(value));
 
     /* Wait for DONE to be set */
     while (!(bcm2835_peri_read_nb(paddr) & BCM2835_SPI0_CS_DONE))
 	;
 
     /* Read any byte that was sent back by the slave while we sere sending to it */
-    ret = bcm2835_peri_read_nb(fifo);
+    ret = bcm2835_correct_order(bcm2835_peri_read_nb(fifo));
 
     /* Set TA = 0, and also set the barrier */
     bcm2835_peri_set_bits(paddr, 0, BCM2835_SPI0_CS_TA);
@@ -652,14 +789,14 @@ void bcm2835_spi_transfernb(char* tbuf, char* rbuf, uint32_t len)
         /* TX fifo not full, so add some more bytes */
         while(((bcm2835_peri_read(paddr) & BCM2835_SPI0_CS_TXD))&&(TXCnt < len ))
         {
-           bcm2835_peri_write_nb(fifo, tbuf[TXCnt]);
-           TXCnt++;
+	    bcm2835_peri_write_nb(fifo, bcm2835_correct_order(tbuf[TXCnt]));
+	    TXCnt++;
         }
         /* Rx fifo not empty, so get the next received bytes */
         while(((bcm2835_peri_read(paddr) & BCM2835_SPI0_CS_RXD))&&( RXCnt < len ))
         {
-           rbuf[RXCnt] = bcm2835_peri_read_nb(fifo);
-           RXCnt++;
+	    rbuf[RXCnt] = bcm2835_correct_order(bcm2835_peri_read_nb(fifo));
+	    RXCnt++;
         }
     }
     /* Wait for DONE to be set */
@@ -671,7 +808,7 @@ void bcm2835_spi_transfernb(char* tbuf, char* rbuf, uint32_t len)
 }
 
 /* Writes an number of bytes to SPI */
-void bcm2835_spi_writenb(char* tbuf, uint32_t len)
+void bcm2835_spi_writenb(const char* tbuf, uint32_t len)
 {
     volatile uint32_t* paddr = bcm2835_spi0 + BCM2835_SPI0_CS/4;
     volatile uint32_t* fifo = bcm2835_spi0 + BCM2835_SPI0_FIFO/4;
@@ -696,7 +833,7 @@ void bcm2835_spi_writenb(char* tbuf, uint32_t len)
 	    ;
 	
 	/* Write to FIFO, no barrier */
-	bcm2835_peri_write_nb(fifo, tbuf[i]);
+	bcm2835_peri_write_nb(fifo, bcm2835_correct_order(tbuf[i]));
 	
 	/* Read from FIFO to prevent stalling */
 	while (bcm2835_peri_read(paddr) & BCM2835_SPI0_CS_RXD)
@@ -736,36 +873,273 @@ void bcm2835_spi_setChipSelectPolarity(uint8_t cs, uint8_t active)
     bcm2835_peri_set_bits(paddr, active << shift, 1 << shift);
 }
 
-int bcm2835_hipi_i2c_begin(int boardrevision)
+void bcm2835_spi_write(uint16_t data)
 {
-    uint16_t cdiv;
-    volatile uint32_t* paddr;
-    
-    if (   bcm2835_bsc0 == MAP_FAILED
-	|| bcm2835_bsc1 == MAP_FAILED)
+#if 0
+	char buf[2];
+
+	buf[0] = data >> 8;
+	buf[1] = data & 0xFF;
+
+	bcm2835_spi_transfern(buf, 2);
+#else
+    volatile uint32_t* paddr = bcm2835_spi0 + BCM2835_SPI0_CS/4;
+    volatile uint32_t* fifo = bcm2835_spi0 + BCM2835_SPI0_FIFO/4;
+
+    /* Clear TX and RX fifos */
+    bcm2835_peri_set_bits(paddr, BCM2835_SPI0_CS_CLEAR, BCM2835_SPI0_CS_CLEAR);
+
+    /* Set TA = 1 */
+    bcm2835_peri_set_bits(paddr, BCM2835_SPI0_CS_TA, BCM2835_SPI0_CS_TA);
+
+	/* Maybe wait for TXD */
+	while (!(bcm2835_peri_read(paddr) & BCM2835_SPI0_CS_TXD))
+	    ;
+
+	/* Write to FIFO */
+	bcm2835_peri_write_nb(fifo,  (uint32_t) data >> 8);
+	bcm2835_peri_write_nb(fifo,  data & 0xFF);
+
+
+    /* Wait for DONE to be set */
+    while (!(bcm2835_peri_read_nb(paddr) & BCM2835_SPI0_CS_DONE))
+	;
+
+    /* Set TA = 0, and also set the barrier */
+    bcm2835_peri_set_bits(paddr, 0, BCM2835_SPI0_CS_TA);
+#endif
+}
+
+int bcm2835_aux_spi_begin(void)
+{
+    volatile uint32_t* enable = bcm2835_aux + BCM2835_AUX_ENABLE/4;
+    volatile uint32_t* cntl0 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL0/4;
+    volatile uint32_t* cntl1 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL1/4;
+
+    if (bcm2835_spi1 == MAP_FAILED)
       return 0; /* bcm2835_init() failed, or not root */
 
-    if( boardrevision == 1 ) {
-        paddr = bcm2835_bsc0 + BCM2835_BSC_DIV/4;
-        /* Set the I2C/BSC0 pins to the Alt 0 function to enable I2C access on them */
-        bcm2835_gpio_fsel(RPI_GPIO_P1_03, BCM2835_GPIO_FSEL_ALT0); /* SDA */
-        bcm2835_gpio_fsel(RPI_GPIO_P1_05, BCM2835_GPIO_FSEL_ALT0); /* SCL */
-    } else {
-        paddr = bcm2835_bsc1 + BCM2835_BSC_DIV/4;
-        /* Set the I2C/BSC1 pins to the Alt 0 function to enable I2C access on them */
-        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_03, BCM2835_GPIO_FSEL_ALT0); /* SDA */
-        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_05, BCM2835_GPIO_FSEL_ALT0); /* SCL */
-    }
+    /* Set the SPI pins to the Alt 4 function to enable SPI1 access on them */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_36, BCM2835_GPIO_FSEL_ALT4);	/* SPI1_CE2_N */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_35, BCM2835_GPIO_FSEL_ALT4);	/* SPI1_MISO */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_38, BCM2835_GPIO_FSEL_ALT4);	/* SPI1_MOSI */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_40, BCM2835_GPIO_FSEL_ALT4);	/* SPI1_SCLK */
 
-    /* Read the clock divider register */
-    cdiv = bcm2835_peri_read(paddr);
-    /* Calculate time for transmitting one byte
-    // 1000000 = micros seconds in a second
-    // 9 = Clocks per byte : 8 bits + ACK
-    */
-    i2c_byte_wait_us = ((float)cdiv / BCM2835_CORE_CLK_HZ) * 1000000 * 9;
+	bcm2835_aux_spi_setClockDivider(bcm2835_aux_spi_CalcClockDivider(1000000));	// Default 1MHz SPI
 
-    return 1;
+	bcm2835_peri_write(enable, BCM2835_AUX_ENABLE_SPI0);
+	bcm2835_peri_write(cntl1, 0);
+	bcm2835_peri_write(cntl0, BCM2835_AUX_SPI_CNTL0_CLEARFIFO);
+
+    return 1; /* OK */
+}
+
+void bcm2835_aux_spi_end(void)
+{
+	/* Set all the SPI1 pins back to input */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_36, BCM2835_GPIO_FSEL_INPT);	/* SPI1_CE2_N */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_35, BCM2835_GPIO_FSEL_INPT);	/* SPI1_MISO */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_38, BCM2835_GPIO_FSEL_INPT);	/* SPI1_MOSI */
+	bcm2835_gpio_fsel(RPI_V2_GPIO_P1_40, BCM2835_GPIO_FSEL_INPT);	/* SPI1_SCLK */
+}
+
+#define DIV_ROUND_UP(n,d)	(((n) + (d) - 1) / (d))
+
+uint16_t bcm2835_aux_spi_CalcClockDivider(uint32_t speed_hz)
+{
+	uint16_t divider;
+
+	if (speed_hz < (uint32_t) BCM2835_AUX_SPI_CLOCK_MIN) {
+		speed_hz = (uint32_t) BCM2835_AUX_SPI_CLOCK_MIN;
+	} else if (speed_hz > (uint32_t) BCM2835_AUX_SPI_CLOCK_MAX) {
+		speed_hz = (uint32_t) BCM2835_AUX_SPI_CLOCK_MAX;
+	}
+
+	divider = (uint16_t) DIV_ROUND_UP(BCM2835_CORE_CLK_HZ, 2 * speed_hz) - 1;
+
+	if (divider > (uint16_t) BCM2835_AUX_SPI_CNTL0_SPEED_MAX) {
+		return (uint16_t) BCM2835_AUX_SPI_CNTL0_SPEED_MAX;
+	}
+
+	return divider;
+}
+
+static uint32_t spi1_speed;
+
+void bcm2835_aux_spi_setClockDivider(uint16_t divider)
+{
+		spi1_speed = (uint32_t) divider;
+}
+
+void bcm2835_aux_spi_write(uint16_t data)
+{
+    volatile uint32_t* cntl0 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL0/4;
+    volatile uint32_t* cntl1 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL1/4;
+    volatile uint32_t* stat = bcm2835_spi1 + BCM2835_AUX_SPI_STAT/4;
+    volatile uint32_t* io = bcm2835_spi1 + BCM2835_AUX_SPI_IO/4;
+
+	uint32_t _cntl0 = (spi1_speed << BCM2835_AUX_SPI_CNTL0_SPEED_SHIFT);
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_CS2_N;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_ENABLE;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_MSBF_OUT;
+	_cntl0 |= 16; // Shift length
+
+	bcm2835_peri_write(cntl0, _cntl0);
+	bcm2835_peri_write(cntl1, BCM2835_AUX_SPI_CNTL1_MSBF_IN);
+
+	while (bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_TX_FULL)
+		;
+
+	bcm2835_peri_write(io, (uint32_t) data << 16);
+}
+
+void bcm2835_aux_spi_writenb(const char *tbuf, uint32_t len) {
+    volatile uint32_t* cntl0 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL0/4;
+    volatile uint32_t* cntl1 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL1/4;
+    volatile uint32_t* stat = bcm2835_spi1 + BCM2835_AUX_SPI_STAT/4;
+    volatile uint32_t* txhold = bcm2835_spi1 + BCM2835_AUX_SPI_TXHOLD/4;
+    volatile uint32_t* io = bcm2835_spi1 + BCM2835_AUX_SPI_IO/4;
+
+	char *tx = (char *) tbuf;
+	uint32_t tx_len = len;
+	uint32_t count;
+	uint32_t data;
+	uint32_t i;
+	uint8_t byte;
+
+	uint32_t _cntl0 = (spi1_speed << BCM2835_AUX_SPI_CNTL0_SPEED_SHIFT);
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_CS2_N;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_ENABLE;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_MSBF_OUT;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_VAR_WIDTH;
+
+	bcm2835_peri_write(cntl0, _cntl0);
+	bcm2835_peri_write(cntl1, BCM2835_AUX_SPI_CNTL1_MSBF_IN);
+
+	while (tx_len > 0) {
+
+		while (bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_TX_FULL)
+			;
+
+		count = MIN(tx_len, 3);
+		data = 0;
+
+		for (i = 0; i < count; i++) {
+			byte = (tx != NULL) ? (uint8_t) *tx++ : (uint8_t) 0;
+			data |= byte << (8 * (2 - i));
+		}
+
+		data |= (count * 8) << 24;
+		tx_len -= count;
+
+		if (tx_len != 0) {
+			bcm2835_peri_write(txhold, data);
+		} else {
+			bcm2835_peri_write(io, data);
+		}
+
+		while (bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_BUSY)
+			;
+
+		(void) bcm2835_peri_read(io);
+	}
+}
+
+void bcm2835_aux_spi_transfernb(const char *tbuf, char *rbuf, uint32_t len) {
+    volatile uint32_t* cntl0 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL0/4;
+    volatile uint32_t* cntl1 = bcm2835_spi1 + BCM2835_AUX_SPI_CNTL1/4;
+    volatile uint32_t* stat = bcm2835_spi1 + BCM2835_AUX_SPI_STAT/4;
+    volatile uint32_t* txhold = bcm2835_spi1 + BCM2835_AUX_SPI_TXHOLD/4;
+    volatile uint32_t* io = bcm2835_spi1 + BCM2835_AUX_SPI_IO/4;
+
+	char *tx = (char *)tbuf;
+	char *rx = (char *)rbuf;
+	uint32_t tx_len = len;
+	uint32_t rx_len = len;
+	uint32_t count;
+	uint32_t data;
+	uint32_t i;
+	uint8_t byte;
+
+	uint32_t _cntl0 = (spi1_speed << BCM2835_AUX_SPI_CNTL0_SPEED_SHIFT);
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_CS2_N;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_ENABLE;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_MSBF_OUT;
+	_cntl0 |= BCM2835_AUX_SPI_CNTL0_VAR_WIDTH;
+
+	bcm2835_peri_write(cntl0, _cntl0);
+	bcm2835_peri_write(cntl1, BCM2835_AUX_SPI_CNTL1_MSBF_IN);
+
+	while ((tx_len > 0) || (rx_len > 0)) {
+
+		while (!(bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_TX_FULL) && (tx_len > 0)) {
+			count = MIN(tx_len, 3);
+			data = 0;
+
+			for (i = 0; i < count; i++) {
+				byte = (tx != NULL) ? (uint8_t) *tx++ : (uint8_t) 0;
+				data |= byte << (8 * (2 - i));
+			}
+
+			data |= (count * 8) << 24;
+			tx_len -= count;
+
+			if (tx_len != 0) {
+				bcm2835_peri_write(txhold, data);
+			} else {
+				bcm2835_peri_write(io, data);
+			}
+
+		}
+
+		while (!(bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_RX_EMPTY) && (rx_len > 0)) {
+			count = MIN(rx_len, 3);
+			data = bcm2835_peri_read(io);
+
+			if (rbuf != NULL) {
+				switch (count) {
+				case 3:
+					*rx++ = (char)((data >> 16) & 0xFF);
+					/*@fallthrough@*/
+					/* no break */
+				case 2:
+					*rx++ = (char)((data >> 8) & 0xFF);
+					/*@fallthrough@*/
+					/* no break */
+				case 1:
+					*rx++ = (char)((data >> 0) & 0xFF);
+				}
+			}
+
+			rx_len -= count;
+		}
+
+		while (!(bcm2835_peri_read(stat) & BCM2835_AUX_SPI_STAT_BUSY) && (rx_len > 0)) {
+			count = MIN(rx_len, 3);
+			data = bcm2835_peri_read(io);
+
+			if (rbuf != NULL) {
+				switch (count) {
+				case 3:
+					*rx++ = (char)((data >> 16) & 0xFF);
+					/*@fallthrough@*/
+					/* no break */
+				case 2:
+					*rx++ = (char)((data >> 8) & 0xFF);
+					/*@fallthrough@*/
+					/* no break */
+				case 1:
+					*rx++ = (char)((data >> 0) & 0xFF);
+				}
+			}
+
+			rx_len -= count;
+		}
+	}
+}
+
+void bcm2835_aux_spi_transfern(char *buf, uint32_t len) {
+	bcm2835_aux_spi_transfernb(buf, buf, len);
 }
 
 int bcm2835_i2c_begin(void)
@@ -812,25 +1186,6 @@ void bcm2835_i2c_end(void)
 #endif
 }
 
-int bcm2835_hipi_i2c_end(int boardrevision)
-{
-    if (   bcm2835_bsc0 == MAP_FAILED
-	|| bcm2835_bsc1 == MAP_FAILED)
-      return 0; /* bcm2835_init() failed, or not root */
-    
-    if( boardrevision == 1 ) {
-        /* Set all the I2C/BSC0 pins back to input */
-        bcm2835_gpio_fsel(RPI_GPIO_P1_03, BCM2835_GPIO_FSEL_INPT); /* SDA */
-        bcm2835_gpio_fsel(RPI_GPIO_P1_05, BCM2835_GPIO_FSEL_INPT); /* SCL */
-    } else {
-        /* Set all the I2C/BSC1 pins back to input */
-        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_03, BCM2835_GPIO_FSEL_INPT); /* SDA */
-        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_05, BCM2835_GPIO_FSEL_INPT); /* SCL */
-    }
-
-    return 1;    
-}
-
 void bcm2835_i2c_setSlaveAddress(uint8_t addr)
 {
     /* Set I2C Device Address */
@@ -842,12 +1197,6 @@ void bcm2835_i2c_setSlaveAddress(uint8_t addr)
     bcm2835_peri_write(paddr, addr);
 }
 
-void bcm2835_hipi_i2c_setSlaveAddress(volatile uint32_t* baseaddress, uint8_t addr )
-{
-   	volatile uint32_t* paddr = baseaddress + BCM2835_BSC_A/4;
-	bcm2835_peri_write(paddr, addr);
-}
-
 /* defaults to 0x5dc, should result in a 166.666 kHz I2C clock frequency.
 // The divisor must be a power of 2. Odd numbers
 // rounded down.
@@ -867,13 +1216,6 @@ void bcm2835_i2c_setClockDivider(uint16_t divider)
     i2c_byte_wait_us = ((float)divider / BCM2835_CORE_CLK_HZ) * 1000000 * 9;
 }
 
-void bcm2835_hipi_i2c_setClockDivider( volatile uint32_t* baseaddress, uint16_t divider )
-{
-    volatile uint32_t* paddr = baseaddress + BCM2835_BSC_DIV/4;
-    bcm2835_peri_write(paddr, divider);
-    i2c_byte_wait_us = ((float)divider / BCM2835_CORE_CLK_HZ) * 1000000 * 9;
-}
-
 /* set I2C clock divider by means of a baudrate number */
 void bcm2835_i2c_set_baudrate(uint32_t baudrate)
 {
@@ -883,14 +1225,6 @@ void bcm2835_i2c_set_baudrate(uint32_t baudrate)
 	bcm2835_i2c_setClockDivider( (uint16_t)divider );
 }
 
-void bcm2835_hipi_i2c_set_baudrate( volatile uint32_t* baseaddress, uint32_t baudrate)
-{
-	uint32_t divider;
-	/* use 0xFFFE mask to limit a max value and round down any odd number */
-	divider = (BCM2835_CORE_CLK_HZ / baudrate) & 0xFFFE;
-	bcm2835_hipi_i2c_setClockDivider( baseaddress, (uint16_t)divider );
-}
-
 /* Writes an number of bytes to I2C */
 uint8_t bcm2835_i2c_write(const char * buf, uint32_t len)
 {
@@ -994,7 +1328,7 @@ uint8_t bcm2835_i2c_read(char* buf, uint32_t len)
     while (!(bcm2835_peri_read_nb(status) & BCM2835_BSC_S_DONE))
     {
         /* we must empty the FIFO as it is populated and not use any delay */
-        while (bcm2835_peri_read_nb(status) & BCM2835_BSC_S_RXD)
+        while (remaining && bcm2835_peri_read_nb(status) & BCM2835_BSC_S_RXD)
     	{
 	    /* Read from FIFO, no barrier */
 	    buf[i] = bcm2835_peri_read_nb(fifo);
@@ -1387,8 +1721,7 @@ int bcm2835_init(void)
 {
     int  memfd;
     int  ok;
-    
-    struct utsname utsbuf;
+    FILE *fp;
     
     if (debug) 
     {
@@ -1402,9 +1735,66 @@ int bcm2835_init(void)
 	bcm2835_bsc0 = bcm2835_peripherals + BCM2835_BSC0_BASE/4;
 	bcm2835_bsc1 = bcm2835_peripherals + BCM2835_BSC1_BASE/4;
 	bcm2835_st   = bcm2835_peripherals + BCM2835_ST_BASE/4;
+	bcm2835_aux  = bcm2835_peripherals + BCM2835_AUX_BASE/4;
+	bcm2835_spi1 = bcm2835_peripherals + BCM2835_SPI1_BASE/4;
+
 	return 1; /* Success */
     }
 
+    /* Figure out the base and size of the peripheral address block
+    // using the device-tree. Required for RPi2/3/4, optional for RPi 1
+    */
+    if ((fp = fopen(BMC2835_RPI2_DT_FILENAME , "rb")))
+    {
+        unsigned char buf[16];
+        uint32_t base_address;
+        uint32_t peri_size;
+        if (fread(buf, 1, sizeof(buf), fp) >= 8)
+        {
+            base_address = (buf[4] << 24) |
+              (buf[5] << 16) |
+              (buf[6] << 8) |
+              (buf[7] << 0);
+            
+            peri_size = (buf[8] << 24) |
+              (buf[9] << 16) |
+              (buf[10] << 8) |
+              (buf[11] << 0);
+            
+            if (!base_address)
+            {
+                /* looks like RPI 4 */
+                base_address = (buf[8] << 24) |
+                      (buf[9] << 16) |
+                      (buf[10] << 8) |
+                      (buf[11] << 0);
+                      
+                peri_size = (buf[12] << 24) |
+                (buf[13] << 16) |
+                (buf[14] << 8) |
+                (buf[15] << 0);
+            }
+            /* check for valid known range formats */
+            if ((buf[0] == 0x7e) &&
+                    (buf[1] == 0x00) &&
+                    (buf[2] == 0x00) &&
+                    (buf[3] == 0x00) &&
+                    ((base_address == BCM2835_PERI_BASE) || (base_address == BCM2835_RPI2_PERI_BASE) || (base_address == BCM2835_RPI4_PERI_BASE)))
+            {
+                bcm2835_peripherals_base = (uint32_t *)base_address;
+                bcm2835_peripherals_size = peri_size;
+                if( base_address == BCM2835_RPI4_PERI_BASE )
+                {
+                    pud_type_rpi4 = 1;
+                }
+            }
+        
+        }
+        
+        fclose(fp);
+    }
+    /* else we are prob on RPi 1 with BCM2835, and use the hardwired defaults */
+
     /* Now get ready to map the peripherals block 
      * If we are not root, try for the new /dev/gpiomem interface and accept
      * the fact that we can only access GPIO
@@ -1412,26 +1802,6 @@ int bcm2835_init(void)
      */
     memfd = -1;
     ok = 0;
-
-    if (uname(&utsbuf) == -1)
-    {
-        fprintf(stderr, "bcm2835_init: Unable to read uname\n") ;
-        goto exit;
-    }
-    
-    if(strncmp(utsbuf.machine, "armv7", 5) == 0)
-    {
-        bcm2835_peripherals_base = (uint32_t *)BCM2836_PERI_BASE;
-    }
-    else if(strncmp(utsbuf.machine, "armv6", 5) == 0)
-    {
-        bcm2835_peripherals_base = (uint32_t *)BCM2835_PERI_BASE;
-    }
-    else {
-        fprintf(stderr, "bcm2835_init: Unknown peripherals base for machine %s\n", utsbuf.machine) ;
-        goto exit;
-    }
-    
     if (geteuid() == 0)
     {
       /* Open the master /dev/mem device */
@@ -1443,7 +1813,7 @@ int bcm2835_init(void)
 	}
       
       /* Base of the peripherals block is mapped to VM */
-      bcm2835_peripherals = mapmem("gpio", bcm2835_peripherals_size, memfd, (uint32_t)bcm2835_peripherals_base);
+      bcm2835_peripherals = mapmem("gpio", bcm2835_peripherals_size, memfd, (off_t)bcm2835_peripherals_base);
       if (bcm2835_peripherals == MAP_FAILED) goto exit;
       
       /* Now compute the base addresses of various peripherals, 
@@ -1458,7 +1828,9 @@ int bcm2835_init(void)
       bcm2835_bsc0 = bcm2835_peripherals + BCM2835_BSC0_BASE/4; /* I2C */
       bcm2835_bsc1 = bcm2835_peripherals + BCM2835_BSC1_BASE/4; /* I2C */
       bcm2835_st   = bcm2835_peripherals + BCM2835_ST_BASE/4;
-      
+      bcm2835_aux  = bcm2835_peripherals + BCM2835_AUX_BASE/4;
+      bcm2835_spi1 = bcm2835_peripherals + BCM2835_SPI1_BASE/4;
+
       ok = 1;
     }
     else
@@ -1474,7 +1846,7 @@ int bcm2835_init(void)
       
       /* Base of the peripherals block is mapped to VM */
       bcm2835_peripherals_base = 0;
-      bcm2835_peripherals = mapmem("gpio", bcm2835_peripherals_size, memfd, (uint32_t)bcm2835_peripherals_base);
+      bcm2835_peripherals = mapmem("gpio", bcm2835_peripherals_size, memfd, (off_t)bcm2835_peripherals_base);
       if (bcm2835_peripherals == MAP_FAILED) goto exit;
       bcm2835_gpio = bcm2835_peripherals;
       ok = 1;
@@ -1505,9 +1877,85 @@ int bcm2835_close(void)
     bcm2835_bsc0 = MAP_FAILED;
     bcm2835_bsc1 = MAP_FAILED;
     bcm2835_st   = MAP_FAILED;
+    bcm2835_aux  = MAP_FAILED;
+    bcm2835_spi1 = MAP_FAILED;
     return 1; /* Success */
 }    
 
+/* HIPI ADDITIONS */
+
+int bcm2835_hipi_i2c_begin(int boardrevision)
+{
+    uint16_t cdiv;
+    volatile uint32_t* paddr;
+    
+    if (   bcm2835_bsc0 == MAP_FAILED
+	|| bcm2835_bsc1 == MAP_FAILED)
+      return 0; /* bcm2835_init() failed, or not root */
+
+    if( boardrevision == 1 ) {
+        paddr = bcm2835_bsc0 + BCM2835_BSC_DIV/4;
+        /* Set the I2C/BSC0 pins to the Alt 0 function to enable I2C access on them */
+        bcm2835_gpio_fsel(RPI_GPIO_P1_03, BCM2835_GPIO_FSEL_ALT0); /* SDA */
+        bcm2835_gpio_fsel(RPI_GPIO_P1_05, BCM2835_GPIO_FSEL_ALT0); /* SCL */
+    } else {
+        paddr = bcm2835_bsc1 + BCM2835_BSC_DIV/4;
+        /* Set the I2C/BSC1 pins to the Alt 0 function to enable I2C access on them */
+        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_03, BCM2835_GPIO_FSEL_ALT0); /* SDA */
+        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_05, BCM2835_GPIO_FSEL_ALT0); /* SCL */
+    }
+
+    /* Read the clock divider register */
+    cdiv = bcm2835_peri_read(paddr);
+    /* Calculate time for transmitting one byte
+    // 1000000 = micros seconds in a second
+    // 9 = Clocks per byte : 8 bits + ACK
+    */
+    i2c_byte_wait_us = ((float)cdiv / BCM2835_CORE_CLK_HZ) * 1000000 * 9;
+
+    return 1;
+}
+
+int bcm2835_hipi_i2c_end(int boardrevision)
+{
+    if (   bcm2835_bsc0 == MAP_FAILED
+	|| bcm2835_bsc1 == MAP_FAILED)
+      return 0; /* bcm2835_init() failed, or not root */
+    
+    if( boardrevision == 1 ) {
+        /* Set all the I2C/BSC0 pins back to input */
+        bcm2835_gpio_fsel(RPI_GPIO_P1_03, BCM2835_GPIO_FSEL_INPT); /* SDA */
+        bcm2835_gpio_fsel(RPI_GPIO_P1_05, BCM2835_GPIO_FSEL_INPT); /* SCL */
+    } else {
+        /* Set all the I2C/BSC1 pins back to input */
+        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_03, BCM2835_GPIO_FSEL_INPT); /* SDA */
+        bcm2835_gpio_fsel(RPI_V2_GPIO_P1_05, BCM2835_GPIO_FSEL_INPT); /* SCL */
+    }
+
+    return 1;    
+}
+
+void bcm2835_hipi_i2c_setSlaveAddress(volatile uint32_t* baseaddress, uint8_t addr )
+{
+   	volatile uint32_t* paddr = baseaddress + BCM2835_BSC_A/4;
+	bcm2835_peri_write(paddr, addr);
+}
+
+void bcm2835_hipi_i2c_setClockDivider( volatile uint32_t* baseaddress, uint16_t divider )
+{
+    volatile uint32_t* paddr = baseaddress + BCM2835_BSC_DIV/4;
+    bcm2835_peri_write(paddr, divider);
+    i2c_byte_wait_us = ((float)divider / BCM2835_CORE_CLK_HZ) * 1000000 * 9;
+}
+
+void bcm2835_hipi_i2c_set_baudrate( volatile uint32_t* baseaddress, uint32_t baudrate)
+{
+	uint32_t divider;
+	/* use 0xFFFE mask to limit a max value and round down any odd number */
+	divider = (BCM2835_CORE_CLK_HZ / baudrate) & 0xFFFE;
+	bcm2835_hipi_i2c_setClockDivider( baseaddress, (uint16_t)divider );
+}
+
 #ifdef BCM2835_TEST
 /* this is a simple test program that prints out what it will do rather than 
 // actually doing it
